"""brain.py provides an example Steno3D project of brain volume"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import os
from six import text_type

from .base import BaseExample, exampleproperty
from ..data import DataArray
from ..project import Project
from ..volume import Mesh3DGrid, Volume


class Brain(BaseExample):
    """Class containing components of brain volume project.

    Construct the project with Brain.get_project()
    """

    @exampleproperty
    def filenames(self):
        """all brain files"""
        bin_files = ['MRbrain.' + text_type(i) for i in range(1, 110)]
        return bin_files + ['info.txt', 'announcement.txt', 'license.txt']

    @exampleproperty
    def datafiles(self):
        """brain data files"""
        filepaths = [
            Brain.fetch_data(filename=fn,
                             download_if_missing=False,
                             verbose=False)
            for fn in Brain.filenames[:-3]
        ]
        return filepaths

    @exampleproperty
    def datavolume(self):
        """return full numpy volume of brain data"""
        dataall = []
        for filename in Brain.datafiles:
            x = np.fromfile(filename, dtype='>i2')
            dataall.append(x.reshape((256, 256), order ='C'))
        return np.array(dataall).astype(float)

    @classmethod
    def get_project(self):
        """return brain volume project"""
        data = Brain.datavolume[::, -30:-210:-2, -30:-210:-2]
        proj = Project(
            title='Brain MRI',
            description='Project with brain volume'
        )
        Volume(
            project=proj,
            mesh=Mesh3DGrid(
                h1=np.ones(data.shape[0]),
                h2=np.ones(data.shape[1]),
                h3=np.ones(data.shape[2]),
            ),
            data=dict(
                location='CC',
                data=DataArray(
                    title='MRI brain data',
                    array=data.flatten()
                )
            ),
            title='Brain Volume'
        )
        return proj



"""surface.py contains the Surface composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps

from numpy import max as npmax
from numpy import min as npmin
from numpy import ndarray
from six import string_types
from traitlets import observe, validate

from .base import BaseMesh
from .base import CompositeResource
from .data import DataArray
from .options import ColorOptions
from .options import MeshOptions
from .texture import Texture2DImage
from .traits import (Array, HasSteno3DTraits, KeywordInstance, Renamed,
                     Repeated, String, Union, Vector)


class _Mesh2DOptions(MeshOptions):
    pass


class _SurfaceOptions(ColorOptions):
    pass


class Mesh2D(BaseMesh):
    """class steno3d.Mesh2D

    Contains spatial information about a 2D surface defined by
    triangular faces.
    """
    vertices = Array(
        help='Mesh vertices',
        shape=('*', 3),
        dtype=float
    )
    triangles = Array(
        help='Mesh triangle vertex indices',
        shape=('*', 3),
        dtype=int
    )
    opts = KeywordInstance(
        help='Mesh2D Options',
        klass=_Mesh2DOptions,
        allow_none=True
    )

    @property
    def nN(self):
        """ get number of nodes """
        return len(self.vertices)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.triangles)

    def _nbytes(self, arr=None):
        if arr is None:
            return self._nbytes('vertices') + self._nbytes('triangles')
        if isinstance(arr, string_types) and arr in ('vertices', 'triangles'):
            arr = getattr(self, arr)
        if isinstance(arr, ndarray):
            return arr.astype('f4').nbytes
        raise ValueError('Mesh2D cannot calculate the number of '
                         'bytes of {}'.format(arr))

    @observe('triangles', 'vertices')
    def _reject_large_files(self, change):
        try:
            self._validate_file_size(change['name'], change['new'])
        except ValueError as err:
            setattr(change['owner'], change['name'], change['old'])
            raise err

    @validate('triangles')
    def _validate_tri(self, proposal):
        if npmin(proposal['value']) < 0:
            raise ValueError('Triangles may only have positive integers')
        if npmax(proposal['value']) >= len(proposal['owner'].vertices):
            raise ValueError('Triangles expects more vertices than provided')
        proposal['owner']._validate_file_size('triangles', proposal['value'])
        return proposal['value']

    @validate('vertices')
    def _validate_vert(self, proposal):
        if npmax(proposal['owner'].triangles) >= len(proposal['value']):
            raise ValueError('Triangles expects more vertices than provided')
        proposal['owner']._validate_file_size('vertices', proposal['value'])
        return proposal['value']

    def _get_dirty_files(self, force=False):
        files = {}
        dirty = self._dirty_traits
        if 'vertices' in dirty or force:
            files['vertices'] = \
                self.traits()['vertices'].serialize(self.vertices)
        if 'triangles' in dirty or force:
            files['triangles'] = \
                self.traits()['triangles'].serialize(self.triangles)
        return files

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        mesh = Mesh2D(
            title=kwargs['title'],
            description=kwargs['description'],
            vertices=Array.download(
                url=json['vertices'],
                shape=(json['verticesSize']//12, 3),
                dtype=json['verticesType']
            ),
            triangles=Array.download(
                url=json['triangles'],
                shape=(json['trianglesSize']//12, 3),
                dtype=json['trianglesType']
            ),
            opts=json['meta']
        )
        return mesh

    @classmethod
    def _build_from_omf(cls, omf_geom, omf_project):
        mesh = Mesh2D(
            vertices=(omf_geom.vertices.array +
                      omf_geom.origin +
                      omf_project.origin),
            triangles=omf_geom.triangles.array
        )
        return mesh


class Mesh2DGrid(BaseMesh):
    """Contains spatial information of a 2D grid."""
    h1 = Array(
        help='Grid cell widths, U-direction',
        shape=('*',),
        dtype=(float, int)
    )
    h2 = Array(
        help='Grid cell widths, V-direction',
        shape=('*',),
        dtype=(float, int)
    )
    x0 = Renamed('O')
    O = Vector(
        help='Origin vector',
        default_value=[0., 0., 0.]
    )
    U = Vector(
        help='Orientation of h1 axis',
        default_value='X'
    )
    V = Vector(
        help='Orientation of h2 axis',
        default_value='Y'
    )
    Z = Array(
        help='Node topography',
        shape=('*',),
        dtype=float,
        default_value=[],
        allow_none=True
    )
    opts = KeywordInstance(
        help='Mesh2D Options',
        klass=_Mesh2DOptions,
        allow_none=True
    )

    @property
    def nN(self):
        """ get number of nodes """
        return (len(self.h1)+1) * (len(self.h2)+1)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.h1) * len(self.h2)

    def _nbytes(self, arr=None):
        filenames = ('h1', 'h2', 'O', 'U', 'V', 'Z')
        if arr is None:
            return sum(self._nbytes(fn) for fn in filenames)
        if isinstance(arr, string_types) and arr in filenames:
            if getattr(self, arr, None) is None:
                return 0
            arr = getattr(self, arr)
        if isinstance(arr, ndarray):
            return arr.astype('f4').nbytes
        raise ValueError('Mesh2DGrid cannot calculate the number of '
                         'bytes of {}'.format(arr))

    @observe('Z')
    def _reject_large_files(self, change):
        try:
            self._validate_file_size(change['name'], change['new'])
        except ValueError as err:
            setattr(change['owner'], change['name'], change['old'])
            raise err

    @validate('Z')
    def _validate_Z(self, proposal):
        """Check if mesh content is built correctly"""
        if len(proposal['value']) == 0:
            return proposal['value']
        if len(proposal['value']) != proposal['owner'].nN:
            raise ValueError(
                'Length of Z, {zlen}, must equal number of nodes, '
                '{nnode}'.format(
                    zlen=len(proposal['value']),
                    nnode=proposal['owner'].nN
                )
            )
        proposal['owner']._validate_file_size('Z', proposal['value'])
        return proposal['value']

    def _get_dirty_data(self, force=False):
        datadict = super(Mesh2DGrid, self)._get_dirty_data(force)
        dirty = self._dirty_traits
        if ('h1' in dirty or 'h2' in dirty) or force:
            datadict['tensors'] = dumps(dict(
                h1=self.h1.tolist(),
                h2=self.h2.tolist(),
            ))
        if ('O' in dirty or 'U' in dirty or 'V' in dirty) or force:
            datadict['OUV'] = dumps(dict(
                O=self.O.tolist(),
                U=Vector.as_length(self.U, self.h1.sum()).tolist(),
                V=Vector.as_length(self.V, self.h2.sum()).tolist()
            ))
        return datadict

    def _get_dirty_files(self, force=False):
        files = super(Mesh2DGrid, self)._get_dirty_files(force)
        dirty = self._dirty_traits
        if 'Z' in dirty or (force and getattr(self, 'Z', []) != []):
            files['Z'] = self.traits()['Z'].serialize(self.Z)
        return files

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        mesh = Mesh2DGrid(
            title=kwargs['title'],
            description=kwargs['description'],
            h1=json['tensors']['h1'],
            h2=json['tensors']['h2'],
            O=json['OUV']['O'],
            U=json['OUV']['U'],
            V=json['OUV']['V'],
            opts=json['meta']
        )
        if json['ZExists']:
            mesh.Z = Array.download(
                url=json['Z'],
                shape=json['ZSize']//4,
                dtype=json['ZType']
            )

        return mesh

    @classmethod
    def _build_from_omf(cls, omf_geom, omf_project):
        mesh = Mesh2DGrid(
            h1=omf_geom.tensor_u,
            h2=omf_geom.tensor_v,
            O=omf_geom.origin + omf_project.origin,
            U=omf_geom.axis_u,
            V=omf_geom.axis_v
        )
        if omf_geom.offset_w is not None:
            mesh.Z = omf_geom.offset_w.array
        return mesh


class _SurfaceBinder(HasSteno3DTraits):
    """Contains the data on a 2D surface with location information"""
    location = String(
        help='Location of the data on mesh',
        choices={
            'CC': ('FACE', 'CELLCENTER'),
            'N': ('NODE', 'VERTEX', 'CORNER')
        }
    )
    data = KeywordInstance(
        help='Data',
        klass=DataArray
    )


class Surface(CompositeResource):
    """Contains all the information about a 2D surface"""
    mesh = Union(
        help='Mesh',
        trait_types=[
            KeywordInstance(klass=Mesh2D),
            KeywordInstance(klass=Mesh2DGrid)
        ]
    )
    data = Repeated(
        help='Data',
        trait=KeywordInstance(klass=_SurfaceBinder),
        allow_none=True
    )
    textures = Repeated(
        help='Textures',
        trait=KeywordInstance(klass=Texture2DImage),
        allow_none=True
    )
    opts = KeywordInstance(
        help='Options',
        klass=_SurfaceOptions,
        allow_none=True
    )

    def _nbytes(self):
        return (self.mesh._nbytes() +
                sum(d.data._nbytes() for d in self.data) +
                sum(t._nbytes() for t in self.textures))

    @validate('data')
    def _validate_data(self, proposal):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(proposal['value']):
            assert dat.location in ('N', 'CC')
            valid_length = (
                proposal['owner'].mesh.nC if dat.location == 'CC'
                else proposal['owner'].mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'surface.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        return proposal['value']


__all__ = ['Surface', 'Mesh2D', 'Mesh2DGrid']



"""vector.py contains the Vector composite resource for Steno3D"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


from traitlets import validate

from .base import CompositeResource
from .options import ColorOptions
from .traits import Array, KeywordInstance, Repeated

from .point import Mesh0D, _PointBinder


class _VectorOptions(ColorOptions):
    pass


class Vector(CompositeResource):
    """Contains all the information about a vector field"""

    _resource_class = 'vector'

    mesh = KeywordInstance(
        help='Mesh',
        klass=Mesh0D
    )
    vectors = Array(
        help='Vector',
        shape=('*', 3),
        dtype=float
    )
    data = Repeated(
        help='Data',
        trait=KeywordInstance(klass=_PointBinder),
        allow_none=True
    )
    opts = KeywordInstance(
        help='Options',
        klass=_VectorOptions,
        allow_none=True
    )

    def _nbytes(self):
        return (self.mesh._nbytes() + self.vectors.astype('f4').nbytes +
                sum(d.data._nbytes() for d in self.data))

    @validate('data')
    def _validate_data(self, proposal):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(proposal['value']):
            assert dat.location == 'N'
            valid_length = proposal['owner'].mesh.nN
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'point.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        return proposal['value']

    @validate('vectors')
    def _validate_vectors(self, proposal):
        """Check if vectors is built correctly"""
        valid_length = proposal['owner'].mesh.nN
        if len(proposal['value']) != valid_length:
            raise ValueError(
                'vectors length {datalen} does not match '
                'mesh vertex length {meshlen}'.format(
                    datalen=len(proposal['value']),
                    meshlen=valid_length
                )
            )
        return proposal['value']

    def _get_dirty_files(self, force=False):
        files = {}
        dirty = self._dirty_traits
        if 'vectors' in dirty or force:
            files['vectors'] = \
                self.traits()['vectors'].serialize(self.vectors)
        return files

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        vec = super(Vector, cls)._build_from_json(json, **kwargs)
        vec.vectors = Array.download(
            url=json['vectors'],
            shape=(json['vectorsSize']//12, 3),
            dtype=json['vectorsType']
        )
        return vec


__all__ = ['Vector']



"""resource.py contains the base resource classes that user-created
resources depend on in steno3d
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
from pprint import pformat
from six import string_types

from traitlets import All, observe, Undefined, validate

from .client import Comms, needs_login, pause, plot
from .traits import (_REGISTRY, HasSteno3DTraits, KeywordInstance, Repeated,
                     String)


class classproperty(property):
    """class decorator to enable property behavior in classmethods"""
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class UserContent(HasSteno3DTraits):
    """Base class for everything user creates and uploads in steno3d"""
    title = String(
        help='Title of the model.',
        default_value='',
        allow_none=True
    )
    description = String(
        help='Description of the model.',
        default_value='',
        allow_none=True
    )
    _sync = False
    _upload_data = None

    @classproperty
    @classmethod
    def _resource_class(cls):
        """name of the class of resource"""
        if getattr(cls, '__resource_class', None) is None:
            cls.__resource_class = cls.__name__.lower()
        return cls.__resource_class

    @classproperty
    @classmethod
    def _model_api_location(cls):
        """api destination for resource"""
        if getattr(cls, '__model_api_location', None) is None:
            cls.__model_api_location = 'resource/{className}'.format(
                className=cls._resource_class
            )
        return cls.__model_api_location

    def _upload(self, sync=False, verbose=True, tab_level=''):
        if getattr(self, '_uploading', False):
            return
        try:
            if verbose:
                print(
                    tab_level + ('Uploading '
                                 if getattr(self, '_upload_data', None) is None
                                 else 'Updating ') +
                    self._resource_class + ': ' + self.title
                )
            self._uploading = True
            pause()
            assert self.validate()
            self._upload_dirty(sync, verbose, tab_level + '    ')
            if getattr(self, '_upload_data', None) is None:
                self._post(
                    self._get_dirty_data(force=True),
                    self._get_dirty_files(force=True)
                )
            else:
                dirty_data = self._get_dirty_data()
                dirty_files = self._get_dirty_files()
                if len(dirty_data) > 0 or len(dirty_files) > 0:
                    self._put(dirty_data, dirty_files)
            self._mark_clean(recurse=False)
            self._sync = sync
            if verbose:
                print(tab_level + '... Complete!')
        except Exception as err:
            if self._sync:
                print('Upload failed, turning off syncing. To restart '
                      'syncing, upload() again.')
                self._sync = False
            else:
                raise err
        finally:
            self._uploading = False

    def _get_dirty_data(self, force=False):
        dirty = self._dirty_traits
        datadict = dict()
        if 'title' in dirty or force:
            datadict['title'] = self.title
        if 'description' in dirty or force:
            datadict['description'] = self.description
        return datadict

    def _get_dirty_files(self, force=False):
        return {}

    def _upload_dirty(self, sync=False, verbose=True, tab_level=''):
        pass

    @observe(All)
    def _on_property_change(self, change):
        if getattr(self, '_sync', False):
            self._upload(self._sync)

    def _post(self, datadict=None, files=None):
        self._client_upload(Comms.post, 'api/' + self._model_api_location,
                            datadict, files)

    def _put(self, datadict=None, files=None):
        pause()
        api_uid = 'api/{mapi}/{uid}'.format(mapi=self._model_api_location,
                                            uid=self._upload_data['uid'])
        self._client_upload(Comms.put, api_uid, datadict, files)

    def _client_upload(self, request_fcn, url,
                       datadict=None, files=None):
        req = request_fcn(
            url,
            data=datadict if datadict else tuple(),
            files=files if files else tuple(),
        )
        if isinstance(req, list):
            for rq in req:
                if rq['status_code'] != 200:
                    try:
                        resp = pformat(rq['json'])
                    except ValueError:
                        resp = rq

                    raise Exception(
                        'Upload failed: {location}'.format(
                            location=url,
                        ) +
                        '\ndata: {datadict}\nfiles: {filedict}'.format(
                            datadict=pformat(datadict),
                            filedict=pformat(files),
                        ) +
                        '\nresponse: {response}'.format(
                            response=resp,
                        )
                    )
            self._upload_data = [rq['json'] for rq in req]
        else:
            if req['status_code'] != 200:
                raise Exception(
                    'Upload failed: {location}'.format(
                        location=url,
                    ) +
                    '\ndata: {datadict}\nfiles: {filedict}'.format(
                        datadict=pformat(datadict),
                        filedict=pformat(files),
                    ) +
                    '\nresponse: {response}'.format(
                        response=req['json'],
                    )
                )
            self._upload_data = req['json']

    @property
    def _json(self):
        """Return a JSON representation of the object"""
        json = getattr(self, '_upload_data', None)
        if json is None:
            raise ValueError('JSON not available: Data not uploaded')
        return json

    @classmethod
    def _json_from_uid(cls, uid):
        if not isinstance(uid, string_types) or len(uid) != 20:
            raise ValueError('{}: invalid uid'.format(uid))
        resp = Comms.get('api/{mapi}/{uid}'.format(
            mapi=cls._model_api_location,
            uid=uid
        ))
        if resp['status_code'] != 200:
            raise ValueError('{uid}: {cls} query failed'.format(
                uid=uid,
                cls=cls._resource_class
            ))
        return resp['json']

    @classmethod
    def _build(cls, src, copy=True, tab_level='', **kwargs):
        if isinstance(src, HasSteno3DTraits):
            raise NotImplementedError('Copying instances not supported')
        print('{tl}Downloading {cls}'.format(
            tl=tab_level,
            cls=cls._resource_class
        ), end=': ')
        if isinstance(src, string_types):
            json = cls._json_from_uid(src)
        else:
            json = src
        title = '' if json['title'] is None else json['title']
        desc = '' if json['description'] is None else json['description']
        print(title)
        res = cls._build_from_json(json, copy=copy, tab_level=tab_level,
                                   title=title, description=desc, **kwargs)
        if not copy:
            res._upload_data = json
        print('{}...Complete!'.format(tab_level))
        return res

    @classmethod
    def _build_from_json(cls, json, copy=True, tab_level='', **kwargs):
        raise NotImplementedError('Cannot build raw UserContent from json')


class BaseResource(UserContent):
    """Base class for all resources that are added to projects and
    uploaded to steno3d
    """

    def _get_dirty_data(self, force=False):
        datadict = super(BaseResource, self)._get_dirty_data(force)
        dirty = self._dirty
        if 'opts' in dirty or (force and hasattr(self, 'opts')):
            datadict['meta'] = self.opts._json
        return datadict


    def _validate_file_size(self, name, arr):
        if Comms.user.logged_in:
            file_limit = Comms.user.file_size_limit
            if self._nbytes(arr) > file_limit:
                raise ValueError(
                    '{name} file size ({file} bytes) exceeds limit: '
                    '{lim} bytes'.format(name=name,
                                         file=self._nbytes(arr),
                                         lim=file_limit)
                )
        return True


class CompositeResource(BaseResource):
    """A composite resource that stores references to lower-level objects."""
    project = Repeated(
        help='Project',
        trait=KeywordInstance(klass='Project')
    )

    def __init__(self, project=None, **kwargs):
        if project is None:
            raise TypeError('Resource must be constructed with its '
                            'containing project(s)')
        super(CompositeResource, self).__init__(**kwargs)
        self.project = project

    @classmethod
    def _url_view_from_uid(cls, uid):
        """Get full url from a uid"""
        url = '{base}{mapi}/{uid}'.format(
            base=Comms.base_url,
            mapi=cls._model_api_location,
            uid=uid)
        return url

    @validate('project')
    def _validate_proj(self, proposal):
        for proj in proposal['value']:
            if self not in proj.resources:
                raise ValueError('Project/resource pointers misaligned: '
                                 'Ensure that projects contain all the '
                                 'resources that point to them.')
        return True

    @needs_login
    def upload(self, sync=False, verbose=True, print_url=True):
        """Upload the resource through its containing project(s)"""
        for proj in self.project:
            proj.upload(sync, verbose, False)
        if print_url:
            print(self._url)
        return self._url



    def _get_dirty_data(self, force=False):
        datadict = super(CompositeResource, self)._get_dirty_data(force)
        dirty = self._dirty_traits
        if 'mesh' in dirty or force:
            datadict['mesh'] = dumps({
                'uid': self.mesh._json['longUid']
            })
        if 'data' in dirty or force:
            datadict['data'] = dumps([
                {
                    'location': d.location,
                    'uid': d.data._json['longUid']
                } for d in self.data
            ])
        if 'textures' in dirty or (force and hasattr(self, 'textures')):
            datadict['textures'] = dumps([
                {
                    'uid': t._json['longUid']
                } for t in self.textures
            ])
        return datadict

    def _upload_dirty(self, sync=False, verbose=True, tab_level=''):
        dirty = self._dirty
        if 'mesh' in dirty:
            self.mesh._upload(sync, verbose, tab_level)
        if 'data' in dirty:
            [d.data._upload(sync, verbose, tab_level) for d in self.data]
        if 'textures' in dirty:
            [t._upload(sync, verbose, tab_level) for t in self.textures]

    @observe('project')
    def _fix_proj_res(self, change):
        before = change['old']
        after = change['new']
        if before in (None, Undefined):
            before = []
        if after in (None, Undefined):
            after = []
        for proj in after:
            if proj not in before and self not in proj.resources:
                proj.resources += [self]
        for proj in before:
            if proj not in after and self in proj.resources:
                proj.resources = [r for r in proj.resources
                                  if r is not self]
        if len(set(after)) != len(after):
            post_post = []
            for p in after:
                if p not in post_post:
                    post_post += [p]
            self.project = post_post

    @property
    def _url(self):
        if getattr(self, '_upload_data', None) is not None:
            return self._url_view_from_uid(self._upload_data['uid'])

    @property
    @needs_login
    def url(self):
        """steno3d.com url of project if uploaded"""
        if getattr(self, '_upload_data', None) is None:
            print('Resource not uploaded: Please upload() '
                  'before accessing the URL.')
        return self._url

    @needs_login
    def plot(self):
        """Display the 3D representation of the content"""
        if getattr(self, '_upload_data', None) is None:
            print('Resource not uploaded: Please upload() '
                  'before plotting.')
            return
        return plot(self._url)

    @classmethod
    def _build_from_json(cls, json, copy=True, tab_level='', **kwargs):
        if 'project' not in kwargs:
            raise KeyError('Building CompositeResource from json requires '
                           'project input.')
        res = cls(
            project=kwargs['project'],
            title=kwargs['title'],
            description=kwargs['description'],
            opts=json['meta']
        )
        (mesh_string, mesh_uid) = (
            json['mesh']['uid'].split('Resource')[-1].split(':')
        )
        mesh_class = _REGISTRY[mesh_string]

        res.mesh = mesh_class._build(mesh_uid, copy, tab_level + '    ')

        if 'textures' in json:
            res.textures = []
            for t in json['textures']:
                (tex_string, tex_uid) = (
                    t['uid'].split('Resource')[-1].split(':')
                )
                tex_class = _REGISTRY[tex_string]
                res.textures += [tex_class._build(
                    tex_uid, copy, tab_level + '    '
                )]

        if 'data' in json:
            res.data = []
            for d in json['data']:
                (data_string, data_uid) = (
                    d['uid'].split('Resource')[-1].split(':')
                )
                data_class = _REGISTRY[data_string]
                res.data += [dict(
                    location=d['location'],
                    data=data_class._build(
                        data_uid, copy, tab_level + '    '
                    )
                )]

        return res

    @classmethod
    def _build_from_omf(cls, omf_element, omf_project, project):
        mesh_map = {
            'PointSetGeometry': 'Mesh0D',
            'LineSetGeometry': 'Mesh1D',
            'SurfaceGeometry': 'Mesh2D',
            'SurfaceGridGeometry': 'Mesh2DGrid',
            'VolumeGridGeometry': 'Mesh3DGrid'
        }
        mesh_class = _REGISTRY[mesh_map[
            omf_element.geometry.__class__.__name__
        ]]
        res = cls(
            project=project,
            title=omf_element.name,
            description=omf_element.description,
            mesh=mesh_class._build_from_omf(omf_element.geometry, omf_project),
            opts={'color': omf_element.color}
        )
        if hasattr(omf_element, 'textures'):
            res.textures = []
            for tex in omf_element.textures:
                res.textures += [
                    _REGISTRY['Texture2DImage']._build_from_omf(tex,
                                                                omf_project)
                ]
        if hasattr(omf_element, 'data'):
            res.data = []
            for dat in omf_element.data:
                if dat.__class__.__name__ not in ('ScalarData', 'MappedData'):
                    print('Data of class {} ignored'.format(
                        dat.__class__.__name__
                    ))
                    continue
                res.data += [
                    _REGISTRY['DataArray']._build_from_omf(dat)
                ]
        return res



class BaseMesh(BaseResource):
    """Base class for all mesh resources. These are contained within
    each composite resources and define its structure
    """


class BaseData(BaseResource):
    """Base class for all data resources. These can be contained within
    each composite resource and define data corresponding to the mesh
    """
    @classproperty
    @classmethod
    def _model_api_location(cls):
        """api destination for texture resource"""
        if getattr(cls, '__model_api_location', None) is None:
            cls.__model_api_location = 'resource/data/{class_name}'.format(
                class_name=cls._resource_class)
        return cls.__model_api_location


class BaseTexture2D(BaseResource):
    """Base class for all texture resources. These can be contained
    within some composite resources and define data in space that gets
    mapped to the mesh.
    """
    @classproperty
    @classmethod
    def _model_api_location(cls):
        """api destination for texture resource"""
        if getattr(cls, '__model_api_location', None) is None:
            cls.__model_api_location = 'resource/texture2d/{cls_name}'.format(
                cls_name=cls._resource_class)
        return cls.__model_api_location



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import namedtuple
from functools import wraps
from io import BytesIO
from tempfile import NamedTemporaryFile
from warnings import warn

import numpy as np
from png import Reader
from requests import get
from six import integer_types, string_types, with_metaclass
import traitlets as tr


_REGISTRY = {}


class MetaDocTraits(tr.MetaHasTraits):

    def __new__(mcs, name, bases, classdict):
        def sphinx(trait_name, trait):
            if isinstance(trait, Steno3DTrait):
                return trait.sphinx(trait_name)
            return (
                ':param {name}: {doc}\n:type {name}: '
                ':class:`{cls} <.{cls}>`'.format(
                    name=trait_name,
                    doc=trait.help,
                    cls=trait.__class__.__name__
                )
            )

        def is_required(trait):
            return not trait.allow_none

        def is_deprecated(trait):
            return isinstance(trait, Renamed)

        def is_optional(trait):
            return not is_required(trait) and not is_deprecated(trait)

        trait_dict = {}
        for base in reversed(bases):
            if issubclass(base, tr.HasTraits):
                trait_dict.update(base.class_traits())

        trait_dict.update({key: value for key, value in classdict.items()
                           if isinstance(value, tr.TraitType)})

        doc_str = classdict.get('__doc__', '')
        req = {key: value for key, value in trait_dict.items()
               if is_required(value)}
        opt = {key: value for key, value in trait_dict.items()
               if is_optional(value)}
        dep = {key: value for key, value in trait_dict.items()
               if is_deprecated(value)}
        if req:
            doc_str += '\n\n**Required**\n\n' + '\n'.join(
                (value.sphinx(key) for key, value in req.items())
            )
        if opt:
            doc_str += '\n\n**Optional**\n\n' + '\n'.join(
                (value.sphinx(key) for key, value in opt.items())
            )
        if dep:
            doc_str += '\n\n**Deprecated**\n\n' + '\n'.join(
                (value.sphinx(key) for key, value in dep.items())
            )
        classdict['__doc__'] = doc_str.strip()

        newcls = super(MetaDocTraits, mcs).__new__(mcs, name, bases, classdict)
        _REGISTRY[name] = newcls
        return newcls


def validator(func):
    """wrapper used on validation functions to recursively validate"""
    @wraps(func)
    def func_wrapper(self):
        if getattr(self, '_validating', False):
            return
        self._cross_validation_lock = False
        self._validating = True
        try:
            trait_dict = self._non_deprecated_traits()
            for k in trait_dict:
                if k in self._trait_values:
                    val = getattr(self, k)
                    trait_dict[k]._validate(self, val)
                    if isinstance(val, DelayedValidator):
                        val.validate()
                    if isinstance(trait_dict[k], Repeated):
                        if len(val) == 0 and not trait_dict[k].allow_none:
                            raise tr.TraitError(
                                'Repeated property must have at least '
                                'one value: {}'.format(k)
                            )
                        for v in val:
                            if isinstance(v, DelayedValidator):
                                v.validate()
                elif not trait_dict[k].allow_none:
                    raise tr.TraitError(
                        'Required property not set: {}'.format(k)
                    )
        finally:
            self._cross_validation_lock = True
            self._validating = False
        return func(self)
    return func_wrapper


class DelayedValidator(tr.HasTraits):

    def __init__(self, *args, **kwargs):
        self._cross_validation_lock = True
        super(DelayedValidator, self).__init__(*args, **kwargs)

    @validator
    def validate(self):
        return True


class HasSteno3DTraits(with_metaclass(MetaDocTraits, DelayedValidator)):

    def __init__(self, **metadata):
        self._dirty_traits = set()
        for key in metadata:
            if key not in self.trait_names():
                raise KeyError('{}: Keyword input is not trait'.format(key))
        super(HasSteno3DTraits, self).__init__(**metadata)

    @tr.observe(tr.All)
    def _mark_dirty(self, change):
        self._dirty_traits.add(change['name'])

    def _mark_clean(self, recurse=True):
        self._dirty_traits = set()
        if not recurse or getattr(self, '_inside_clean', False):
            return
        self._inside_clean = True
        try:
            traits = self._dirty
            for trait in traits:
                value = getattr(self, trait)
                if isinstance(value, HasSteno3DTraits):
                    value._mark_clean()
                if isinstance(value, (list, tuple)):
                    for v in value:
                        if isinstance(v, HasSteno3DTraits):
                            v._mark_clean()
        finally:
            self._inside_clean = False

    @property
    def _dirty(self):
        if getattr(self, '_inside_dirty', False):
            return set()
        dirty_instances = set()
        self._inside_dirty = True
        try:
            traits = self._non_deprecated_traits()
            for trait in traits:
                value = getattr(self, trait)
                if (isinstance(value, HasSteno3DTraits) and
                        len(value._dirty) > 0):
                    dirty_instances.add(trait)
                if isinstance(value, (list, tuple)):
                    for v in value:
                        if (isinstance(v, HasSteno3DTraits) and
                                len(v._dirty) > 0):
                            dirty_instances.add(trait)
        finally:
            self._inside_dirty = False
        return self._dirty_traits.union(dirty_instances)

    def _non_deprecated_traits(self):
        return {k: v for k, v in self.traits().items()
                if not isinstance(v, Renamed)}



class Steno3DTrait(object):
    """A mixin for identifying and documenting steno3d traits"""

    sphinx_extra = ''

    @property
    def sphinx_class(self):
        return ':class:`{cls} <.{cls}>`'.format(cls=self.__class__.__name__)

    def sphinx(self, name):
        if not isinstance(self, tr.TraitType):
            return ''
        return (
            ':param {name}: {doc}{default}\n:type {name}: {cls}'.format(
                name=name,
                doc=self.help + self.sphinx_extra,
                default=('' if (self.default_value is tr.Undefined or
                                self.default_value is None or
                                self.default_value in ([], {}, ''))
                         else ', Default: ' + str(self.default_value)),
                cls=self.sphinx_class
            )
        )


class Steno3DNumber(Steno3DTrait):

    @property
    def sphinx_extra(self):
        if (getattr(self, 'min', None) is None and
                getattr(self, 'max', None) is None):
            return ''
        return ', Range: [{mn}, {mx}]'.format(
            mn='-inf' if getattr(self, 'min', None) is None else self.min,
            mx='inf' if getattr(self, 'max', None) is None else self.max
        )


class Int(Steno3DNumber, tr.Int):
    pass


class Float(Steno3DNumber, tr.Float):
    pass


class Bool(Steno3DTrait, tr.Bool):
    pass


class Union(Steno3DTrait, tr.Union):

    @property
    def sphinx_class(self):
        for t in self.trait_types:
            if not isinstance(t, Steno3DTrait):
                return super(Union, self).sphinx_class
        return ', '.join(t.sphinx_class for t in self.trait_types)

    def info(self):
        return " or ".join([tt.info() for tt in self.trait_types])


class Color(Steno3DTrait, tr.TraitType):
    """A trait for rgb, hex, or string colors. Converts to rgb."""

    default_value = 'RANDOM'
    info_text = ('a color (RGB with values 0-255, hex color e.g. \'#FF0000\', '
                 'string color name, or \'random\')')
    sphinx_extra = ', Format: RGB, hex, or predefined color'

    def validate(self, obj, value):
        """check if input is valid color and converts to RBG"""
        if isinstance(value, string_types):
            if value in COLORS_NAMED:
                value = COLORS_NAMED[value]
            if value.upper() == 'RANDOM':
                value = COLORS_20[np.random.randint(0, 20)]
            value = value.upper().lstrip('#')
            if len(value) == 3:
                value = ''.join(v*2 for v in value)
            if len(value) != 6:
                self.error(obj, value)
            try:
                value = [
                    int(value[i:i + 6 // 3], 16) for i in range(0, 6, 6 // 3)
                ]
            except ValueError:
                self.error(obj, value)
        if not isinstance(value, (list, tuple)):
            self.error(obj, value)
        if len(value) != 3:
            self.error(obj, value)
        for v in value:
            if not isinstance(v, integer_types) or not 0 <= v <= 255:
                self.error(obj, value)
        return tuple(value)


class String(Steno3DTrait, tr.TraitType):
    """A trait for strings, where you can map several values to one"""

    def __init__(self, choices=None, lowercase=False, strip=' ',
                 default_value=tr.Undefined, **metadata):
        if choices is None:
            choices = {}
        if not isinstance(choices, (list, tuple, dict)):
            raise tr.TraitError('choices must be a list, tuple, or dict')
        if isinstance(choices, (list, tuple)):
            choices = {v: v for v in choices}
        for k, v in choices.items():
            if not isinstance(v, (list, tuple)):
                choices[k] = [v]
        for k, v in choices.items():
            if not isinstance(k, string_types):
                raise tr.TraitError('choices must be strings')
            for val in v:
                if not isinstance(val, string_types):
                    raise tr.TraitError('choices must be strings')
        self.lowercase = lowercase
        self.strip = strip
        self.choices = choices
        if metadata.get('allow_none', False) and default_value is tr.Undefined:
            default_value = None
        super(String, self).__init__(default_value=default_value, **metadata)

    def validate(self, obj, value):
        """check that input is string and in choices, if applicable"""
        if not isinstance(value, string_types):
            self.error(obj, value)
        if self.strip is not None:
            value = value.strip(self.strip)
        if self.choices is not None and len(self.choices) != 0:
            for k, v in self.choices.items():
                if (value.upper() == k.upper() or
                        value.upper() in [_.upper() for _ in v]):
                    return k.lower() if self.lowercase else k
            self.error(obj, value)
        return value.lower() if self.lowercase else value

    def info(self):
        """description of string trait"""
        if self.choices is not None and len(self.choices) != 0:
            result = 'any of ' + ', '.join(self.choices)
        else:
            result = 'a string'
        if self.allow_none:
            return result + ' or None'
        return result

    @property
    def sphinx_extra(self):
        if self.choices is None or len(self.choices) == 0:
            return ''
        return ', Choices: ' + ', '.join(self.choices)


class Image(Steno3DTrait, tr.TraitType):
    """A trait for PNG images"""

    info_text = 'a PNG image file'
    sphinx_extra = ', Format: PNG'

    def validate(self, obj, value):
        """checks that image file is PNG and gets a copy"""
        if getattr(value, '__valid__', False):
            return value

        try:
            if hasattr(value, 'read'):
                Reader(value).validate_signature()
            else:
                with open(value, 'rb') as v:
                    Reader(v).validate_signature()
        except Exception:
            self.error(obj, value)

        output = BytesIO()
        output.name = 'texture.png'
        output.__valid__ = True
        if hasattr(value, 'read'):
            fp = value
            fp.seek(0)
        else:
            fp = open(value, 'rb')
        output.write(fp.read())
        output.seek(0)
        fp.close()
        return output

    @classmethod
    def download(cls, url):
        im_resp = get(url)
        if im_resp.status_code != 200:
            raise IOError('Failed to download image.')
        output = BytesIO()
        output.name = 'texture.png'
        for chunk in im_resp:
            output.write(chunk)
        output.seek(0)
        return output


FileProp = namedtuple('FileProp', ['file', 'dtype'])


class Array(Steno3DTrait, tr.TraitType):
    """A trait for serializable float or int arrays"""

    def __init__(self, shape=('*',), dtype=(float, int), **metadata):
        if not isinstance(shape, tuple):
            raise tr.TraitError("{}: Invalid shape - must be a tuple "
                                "(e.g. ('*',3) for an array of length-3 "
                                "arrays)".format(shape))
        for s in shape:
            if s != '*' and not isinstance(s, integer_types):
                raise tr.TraitError("{}: Invalid shape - values "
                                    "must be '*' or int".format(shape))
        self.shape = shape

        if not isinstance(dtype, (list, tuple)):
            dtype = (dtype,)
        if (float not in dtype and
                len(set(dtype).intersection(integer_types)) == 0):
            raise tr.TraitError("{}: Invalid dtype - must be int "
                                "and/or float".format(dtype))
        self.dtype = dtype
        super(Array, self).__init__(**metadata)

    def info(self):
        return 'a list or numpy array of {type} with shape {shp}'.format(
            type=', '.join([str(t) for t in self.dtype]),
            shp=self.shape
        )

    @property
    def sphinx_extra(self):
        return ', Shape: {shp}, Type: {dtype}'.format(
            shp='(' + ', '.join(['\*' if s == '*' else str(s)
                                 for s in self.shape]) + ')',
            dtype=self.dtype
        )

    def validate(self, obj, value):
        """Determine if array is valid based on shape and dtype"""
        if not isinstance(value, (list, np.ndarray)):
            self.error(obj, value)
        value = np.array(value)
        if (value.dtype.kind == 'i' and
                len(set(self.dtype).intersection(integer_types)) == 0):
            self.error(obj, value)
        if value.dtype.kind == 'f' and float not in self.dtype:
            self.error(obj, value)
        if len(self.shape) != value.ndim:
            self.error(obj, value)
        for i, s in enumerate(self.shape):
            if s != '*' and value.shape[i] != s:
                self.error(obj, value)
        return value

    def serialize(self, data):
        """Convert the array data to a serialized binary format"""
        if isinstance(data.flatten()[0], np.floating):
            use_dtype = '<f4'
            nan_mask = ~np.isnan(data)
            assert np.allclose(
                    data.astype(use_dtype)[nan_mask], data[nan_mask]), \
                'Converting the type should not screw things up.'
        elif isinstance(data.flatten()[0], np.integer):
            use_dtype = '<i4'
            assert (data.astype(use_dtype) == data).all(), \
                'Converting the type should not screw things up.'
        else:
            raise Exception('Must be a float or an int: {}'.format(data.dtype))

        data_file = NamedTemporaryFile('rb+', suffix='.dat')
        data.astype(use_dtype).tofile(data_file.file)
        data_file.seek(0)
        return FileProp(data_file, use_dtype)

    @classmethod
    def download(cls, url, shape, dtype=float):
        arr_resp = get(url)
        if arr_resp.status_code != 200:
            raise IOError('Failed to download array.')
        data_file = NamedTemporaryFile()
        for chunk in arr_resp:
            data_file.write(chunk)
        data_file.seek(0)
        arr = np.fromfile(data_file.file, dtype).reshape(shape)
        data_file.close()
        return arr


class Vector(Array):
    """A trait for 3D vectors

    Must be a length-3 array. Int vectors are casted to Float vectors.
    X, Y, and Z are also valid inputs. These correspond to [1., 0, 0],
    [0, 1., 0], and [0, 0, 1.] respectively
    """

    def __init__(self, **metadata):
        super(Vector, self).__init__(
            shape=(3,), dtype=(float, int), **metadata
        )

    def validate(self, obj, value):
        if isinstance(value, string_types):
            if value.upper() == 'X':
                value = [1., 0, 0]
            elif value.upper() == 'Y':
                value = [0., 1, 0]
            elif value.upper() == 'Z':
                value = [0., 0, 1]
        if not isinstance(value, (list, np.ndarray)):
            self.error(obj, value)
        value = np.array(value)
        if value.dtype.kind not in ('f', 'i'):
            self.error(obj, value)
        value = value.astype('float')
        if value.ndim == 2 and value.shape[0] == 1:
            value = value[0]
        if value.ndim != 1:
            self.error(obj, value)
        if len(value) != 3:
            self.error(obj, value)
        return value

    @staticmethod
    def as_length(vector, new_length):
        length = np.sqrt(np.sum(vector**2))
        if length == 0:
            raise ZeroDivisionError('Cannot resize vector of length 0')
        return new_length*vector/length

    @staticmethod
    def normalize(vector):
        return Vector.as_length(vector, 1)


class KeywordInstance(Steno3DTrait, tr.Instance):
    """An instance trait that can be constructed with only a keyword dict"""

    def __init__(self, klass=None, args=(), kw=None, **metadata):
        if klass is None:
            raise tr.TraitError('KeywordInstance klass cannot be None')
        super(KeywordInstance, self).__init__(klass, args, kw, **metadata)

    def _resolve_string(self, string):
        try:
            return _REGISTRY[string]
        except KeyError:
            return super(KeywordInstance, self)._resolve_string(string)

    def validate(self, obj, value):
        if isinstance(value, self.klass):
            return value
        if isinstance(value, dict):
            try:
                return self.klass(**value)
            except:
                self.error(obj, value)
        try:
            return self.klass(value)
        except:
            self.error(obj, value)

    def info(self):
        if isinstance(self.klass, string_types):
            klass = self.klass
        else:
            klass = self.klass.__name__
        result = tr.class_of(klass)
        result = result + ' or a keyword dictionary to construct ' + result
        if self.allow_none:
            return result + ' or None'
        return result

    @property
    def sphinx_class(self):
        if isinstance(self.klass, string_types):
            kls = self.klass
        else:
            kls = self.klass.__name__
        return ':class:`{cls} <.{cls}>`'.format(cls=kls)


class Repeated(Steno3DTrait, tr.List):
    """A list trait that creates a length-1 list if given an instance"""

    @property
    def sphinx_class(self):
        if self._trait is None or not isinstance(self._trait, Steno3DTrait):
            return super(Repeated, self).sphinx_class
        return self._trait.sphinx_class

    @property
    def sphinx_extra(self):
        message = ' (Multiple values may be specified using a list)'
        if self._trait is not None and isinstance(self._trait, Steno3DTrait):
            return self._trait.sphinx_extra + message
        return message

    def get(self, obj, cls=None):
        value = super(Repeated, self).get(obj, cls)
        return [v for v in value]

    def validate(self, obj, value):
        if not isinstance(value, (list, tuple)):
            value = [value]
        return super(Repeated, self).validate(obj, value)


class Renamed(Steno3DTrait, tr.TraitType):
    """For renamed traits, warns and reassigns"""

    def __init__(self, new_name, **metadata):
        self.new_name = new_name
        super(Renamed, self).__init__(allow_none=True, **metadata)

    @property
    def sphinx_class(self):
        return ''

    @property
    def sphinx_extra(self):
        return ('This trait is deprecated and may be removed in the future. '
                'Please use **{}** instead.'.format(self.new_name))

    def _warn(self):
        warn(
            '\n`{}` trait deprecated and may be removed in the future. '
            'Please use `{}`.'.format(self.name, self.new_name),
            FutureWarning, stacklevel=3
        )

    def get(self, obj, cls=None):
        self._warn()
        return getattr(obj, self.new_name)

    def set(self, obj, value):
        self._warn()
        return setattr(obj, self.new_name, value)


COLORS_20 = [
    '#1f77b4', '#aec7e8', '#ff7f0e', '#ffbb78', '#2ca02c',
    '#98df8a', '#d62728', '#ff9896', '#9467bd', '#c5b0d5',
    '#8c564b', '#c49c94', '#e377c2', '#f7b6d2', '#7f7f7f',
    '#c7c7c7', '#bcbd22', '#dbdb8d', '#17becf', '#9edae5'
]

COLORS_NAMED = dict(
    aliceblue="F0F8FF", antiquewhite="FAEBD7", aqua="00FFFF",
    aquamarine="7FFFD4", azure="F0FFFF", beige="F5F5DC",
    bisque="FFE4C4", black="000000", blanchedalmond="FFEBCD",
    blue="0000FF", blueviolet="8A2BE2", brown="A52A2A",
    burlywood="DEB887", cadetblue="5F9EA0", chartreuse="7FFF00",
    chocolate="D2691E", coral="FF7F50", cornflowerblue="6495ED",
    cornsilk="FFF8DC", crimson="DC143C", cyan="00FFFF",
    darkblue="00008B", darkcyan="008B8B", darkgoldenrod="B8860B",
    darkgray="A9A9A9", darkgrey="A9A9A9", darkgreen="006400",
    darkkhaki="BDB76B", darkmagenta="8B008B", darkolivegreen="556B2F",
    darkorange="FF8C00", darkorchid="9932CC", darkred="8B0000",
    darksalmon="E9967A", darkseagreen="8FBC8F", darkslateblue="483D8B",
    darkslategray="2F4F4F", darkslategrey="2F4F4F", darkturquoise="00CED1",
    darkviolet="9400D3", deeppink="FF1493", deepskyblue="00BFFF",
    dimgray="696969", dimgrey="696969", dodgerblue="1E90FF",
    irebrick="B22222", floralwhite="FFFAF0", forestgreen="228B22",
    fuchsia="FF00FF", gainsboro="DCDCDC", ghostwhite="F8F8FF",
    gold="FFD700", goldenrod="DAA520", gray="808080",
    grey="808080", green="008000", greenyellow="ADFF2F",
    honeydew="F0FFF0", hotpink="FF69B4", indianred="CD5C5C",
    indigo="4B0082", ivory="FFFFF0", khaki="F0E68C",
    lavender="E6E6FA", lavenderblush="FFF0F5", lawngreen="7CFC00",
    lemonchiffon="FFFACD", lightblue="ADD8E6", lightcoral="F08080",
    lightcyan="E0FFFF", lightgoldenrodyellow="FAFAD2", lightgray="D3D3D3",
    lightgrey="D3D3D3", lightgreen="90EE90", lightpink="FFB6C1",
    lightsalmon="FFA07A", lightseagreen="20B2AA", lightskyblue="87CEFA",
    lightslategray="778899", lightslategrey="778899", lightsteelblue="B0C4DE",
    lightyellow="FFFFE0", lime="00FF00", limegreen="32CD32",
    linen="FAF0E6", magenta="FF00FF", maroon="800000",
    mediumaquamarine="66CDAA", mediumblue="0000CD", mediumorchid="BA55D3",
    mediumpurple="9370DB", mediumseagreen="3CB371", mediumslateblue="7B68EE",
    mediumspringgreen="00FA9A", mediumturquoise="48D1CC",
    mediumvioletred="C71585", midnightblue="191970", mintcream="F5FFFA",
    mistyrose="FFE4E1", moccasin="FFE4B5", navajowhite="FFDEAD",
    navy="000080", oldlace="FDF5E6", olive="808000",
    olivedrab="6B8E23", orange="FFA500", orangered="FF4500",
    orchid="DA70D6", palegoldenrod="EEE8AA", palegreen="98FB98",
    paleturquoise="AFEEEE", palevioletred="DB7093", papayawhip="FFEFD5",
    peachpuff="FFDAB9", peru="CD853F", pink="FFC0CB",
    plum="DDA0DD", powderblue="B0E0E6", purple="800080",
    rebeccapurple="663399", red="FF0000", rosybrown="BC8F8F",
    royalblue="4169E1", saddlebrown="8B4513", salmon="FA8072",
    sandybrown="F4A460", seagreen="2E8B57", seashell="FFF5EE",
    sienna="A0522D", silver="C0C0C0", skyblue="87CEEB",
    slateblue="6A5ACD", slategray="708090", slategrey="708090",
    snow="FFFAFA", springgreen="00FF7F", steelblue="4682B4",
    tan="D2B48C", teal="008080", thistle="D8BFD8",
    tomato="FF6347", turquoise="40E0D0", violet="EE82EE",
    wheat="F5DEB3", white="FFFFFF", whitesmoke="F5F5F5",
    yellow="FFFF00", yellowgreen="9ACD32"
)



"""Welcome to the Python client library for Steno3D!

This library is used to construct 3D projects and upload them to
steno3d.com - if you do not have an account you can go to
steno3d.com/signup and sign up. You will also need to request a developer
API key at steno3d.com/settings/developer. Online documentation for this
python client is found at steno3d.com/docs.

Steno3D is built with tab-completion in mind; if you are in an
interacitve environment such as a Jupyter Notebook it is easy to explore
the base namespace and seek additional documentation and help.

If you would like to get started with some example resources,
`from steno3d import examples`
should provide a few pre-made examples to look at.

When you are ready to upload your project to steno3d.com, then
`steno3d.login()`
and provide your developer API key.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from atexit import register

from . import parsers
from . import query
from . import client
from .project import *
from .data import *
from .line import *
from .point import *
from .surface import *
from .texture import *
from .vector import *
from .volume import *

__version__ = '0.2.14'
__author__ = '3point Science'
__license__ = 'MIT'
__copyright__ = 'Copyright 2016 3point Science'

login = client.Comms.login
logout = client.Comms.logout
# register(logout)

try:
    del project, data, line, point, surface, texture, traits, volume
    del base, client, options, user
    del absolute_import, division, print_function, unicode_literals
    del register
except NameError:
    # Error cleaning namespace
    pass



"""tsyganenko.py provides an example Steno3D project of mag field lines"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np

from .base import BaseExample, exampleproperty
from ..data import DataArray
from ..line import Mesh1D, Line
from ..project import Project


class Tsyganenko(BaseExample):
    """Class containing components of magnetic field lines project.

    Construct the project with Tsyganenko.get_project()
    """

    @exampleproperty
    def filenames(self):
        """all Tsyganenko files"""
        return ['tsyganenko_mag_field.txt', 'license.txt']

    @exampleproperty
    def datafile(self):
        """Tsyganenko data file"""
        return Tsyganenko.fetch_data(filename='tsyganenko_mag_field.txt',
                                     download_if_missing=False,
                                     verbose=False)

    @staticmethod
    def read_file(fname):
        verts = []
        segs = []
        data = []
        with open(fname, 'r') as f:
            for line in f:
                items = line.split()
                verts.append([float(i) for i in items[1:4]])
                data.append(float(items[4]))
                if items[0] != '1':
                    segs.append([len(verts)-2, len(verts)-1])
        return verts, segs, data

    @classmethod
    def get_project(self):
        """return rocket engine surface project"""
        verts, segs, data = Tsyganenko.read_file(Tsyganenko.datafile)
        proj = Project(
            title='Tsyganenko Magnetic Field',
            description='Simple model of the magnetic field of the Earth'
        )
        Line(
            project=proj,
            mesh=Mesh1D(
                vertices=verts,
                segments=segs,
            ),
            data=[
                dict(
                    location='N',
                    data=DataArray(
                        title='Magnitude of magnetic field (nT)',
                        array=data
                    )
                ),
                dict(
                    location='N',
                    data=DataArray(
                        title='Natural log of field (log(nT))',
                        array=np.log(data)
                    )
                )
            ],
            title='Magnetic Field Lines'
        )
        return proj



"""volume.py contains the Volume composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
from numpy import ndarray
from six import string_types
from traitlets import validate

from .base import BaseMesh
from .base import CompositeResource
from .data import DataArray
from .options import ColorOptions
from .options import MeshOptions
from .traits import (Array, HasSteno3DTraits, KeywordInstance, Renamed,
                     Repeated, String, Vector)


class _Mesh3DOptions(MeshOptions):
    pass


class _VolumeOptions(ColorOptions):
    pass


class Mesh3DGrid(BaseMesh):
    """Contains spatial information of a 3D grid volume."""

    h1 = Array(
        help='Tensor cell widths, x-direction',
        shape=('*',),
        dtype=(float, int)
    )
    h2 = Array(
        help='Tensor cell widths, y-direction',
        shape=('*',),
        dtype=(float, int)
    )
    h3 = Array(
        help='Tensor cell widths, z-direction',
        shape=('*',),
        dtype=(float, int)
    )
    x0 = Renamed('O')
    O = Vector(
        help='Origin vector',
        default_value=[0., 0., 0.]
    )
    U = Vector(
        help='Orientation of h1 axis',
        default_value='X'
    )
    V = Vector(
        help='Orientation of h2 axis',
        default_value='Y'
    )
    W = Vector(
        help='Orientation of h3 axis',
        default_value='Z'
    )
    opts = KeywordInstance(
        help='Mesh3D Options',
        klass=_Mesh3DOptions,
        allow_none=True
    )

    @property
    def nN(self):
        """ get number of nodes """
        return (len(self.h1)+1) * (len(self.h2)+1) * (len(self.h3)+1)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.h1) * len(self.h2) * len(self.h3)

    def _nbytes(self, arr=None):
        filenames = ('h1', 'h2', 'h3', 'O', 'U', 'V', 'W')
        if arr is None:
            return sum(self._nbytes(fn) for fn in filenames)
        if isinstance(arr, string_types) and arr in filenames:
            if getattr(self, arr, None) is None:
                return 0
            arr = getattr(self, arr)
        if isinstance(arr, ndarray):
            return arr.astype('f4').nbytes
        raise ValueError('Mesh3DGrid cannot calculate the number of '
                         'bytes of {}'.format(arr))

    def _get_dirty_data(self, force=False):
        datadict = super(Mesh3DGrid, self)._get_dirty_data(force)
        dirty = self._dirty_traits
        if force or ('h1' in dirty or 'h2' in dirty or 'h3' in dirty):
            datadict['tensors'] = dumps(dict(
                h1=self.h1.tolist(),
                h2=self.h2.tolist(),
                h3=self.h3.tolist()
            ))
        if force or any([item in dirty for item in
                         ['O', 'U', 'V', 'W', 'h1', 'h2', 'h3']]):
            datadict['OUVZ'] = dumps(dict(
                O=self.O.tolist(),
                U=Vector.as_length(self.U, self.h1.sum()).tolist(),
                V=Vector.as_length(self.V, self.h2.sum()).tolist(),
                Z=Vector.as_length(self.W, self.h3.sum()).tolist()
            ))
        return datadict

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        mesh = Mesh3DGrid(
            title=kwargs['title'],
            description=kwargs['description'],
            h1=json['tensors']['h1'],
            h2=json['tensors']['h2'],
            h3=json['tensors']['h3'],
            O=json['OUVZ']['O'],
            U=json['OUVZ']['U'],
            V=json['OUVZ']['V'],
            W=json['OUVZ']['Z'],
            opts=json['meta']
        )
        return mesh

    @classmethod
    def _build_from_omf(cls, omf_geom, omf_project):
        mesh = Mesh3DGrid(
            h1=omf_geom.tensor_u,
            h2=omf_geom.tensor_v,
            h3=omf_geom.tensor_w,
            O=omf_geom.origin + omf_project.origin,
            U=omf_geom.axis_u,
            V=omf_geom.axis_v,
            W=omf_geom.axis_w
        )
        return mesh


class _VolumeBinder(HasSteno3DTraits):
    """Contains the data on a 3D volume with location information"""
    location = String(
        help='Location of the data on mesh',
        choices={
            'CC': ('CELLCENTER'),
            # 'N': ('NODE', 'VERTEX', 'CORNER')
        }
    )
    data = KeywordInstance(
        help='Data',
        klass=DataArray
    )


class Volume(CompositeResource):
    """Contains all the information about a 3D volume"""
    mesh = KeywordInstance(
        help='Mesh',
        klass=Mesh3DGrid,
    )
    data = Repeated(
        help='Data',
        trait=KeywordInstance(klass=_VolumeBinder)
    )
    opts = KeywordInstance(
        help='Options',
        klass=_VolumeOptions,
        allow_none=True
    )

    def _nbytes(self):
        return self.mesh._nbytes() + sum(d.data._nbytes() for d in self.data)

    @validate('data')
    def _validate_data(self, proposal):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(proposal['value']):
            assert dat.location == 'CC'  # in ('N', 'CC')
            valid_length = (
                proposal['owner'].mesh.nC if dat.location == 'CC'
                else proposal['owner'].mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'volume.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        return proposal['value']


__all__ = ['Volume', 'Mesh3DGrid']



"""project.py contains the project class that contains resources
in steno3d
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six

from traitlets import observe, Undefined, validate

from .base import CompositeResource, UserContent
from .client import Comms, needs_login, plot
from .traits import _REGISTRY, Bool, KeywordInstance, Repeated


QUOTA_REACHED = """
Uploading this {priv} project will put you over your quota
of {num} {priv} project(s). For more projects and space, consider
upgrading your account: https://steno3d.com/settings/plan
"""

QUOTA_IMPENDING = """
After this project, you may upload {remaining} more {priv} project(s) before
reaching your {priv} project quota. For more projects and space
consider upgrading your account: https://steno3d.com/settings/plan
"""


class Project(UserContent):
    """Steno3D top-level project"""
    _model_api_location = 'project/steno3d'

    resources = Repeated(
        help='Project Resources',
        trait=KeywordInstance(klass=CompositeResource)
    )

    public = Bool(
        help='Public visibility of project',
        default_value=False
    )

    _public_online = None

    @classmethod
    def _url_view_from_uid(cls, uid):
        """Get full url from a uid"""
        url = '{base}{mapi}/{uid}'.format(
            base=Comms.base_url,
            mapi='app',
            uid=uid)
        return url

    @needs_login
    def upload(self, sync=False, verbose=True, print_url=True):
        """Upload the project"""
        if getattr(self, '_upload_data', None) is None:
            assert self.validate()

            self._check_project_quota(verbose)
            self._public_online = self.public
        elif verbose and self._public_online:
            print('This project is PUBLIC. It is viewable by everyone.')
        if verbose and not self._public_online == self.public:
            print('Local privacy changes cannot be applied to '
                  'projects that are already uploaded. To make '
                  'these changes, please use the dashboard on '
                  'steno3d.com.')
        self._upload(sync, verbose)
        self._trigger_ACL_fix()
        if print_url:
            print(self._url)
        return self._url

    def _trigger_ACL_fix(self):
        self._put({})

    @validate('resources')
    def _validate_resources(self, proposal):
        """Check if project resource pointers are correct"""
        for res in proposal['value']:
            if self not in res.project:
                raise ValueError('Project/resource pointers misaligned: '
                                 'Ensure that resources point to containing '
                                 'project.')
        self._validate_project_size()
        return True

    def _validate_project_size(self, res=None):
        if res is None:
            res = self.resources
        if Comms.user.logged_in:
            res_limit = Comms.user.project_resource_limit
            if len(res) > res_limit:
                raise ValueError(
                    'Total number of resources in project ({res}) '
                    'exceeds limit: {lim}'.format(res=len(self.resources),
                                                  lim=res_limit)
                )
            size_limit = Comms.user.project_size_limit
            sz = sum(r._nbytes() for r in res)
            if sz > size_limit:
                raise ValueError(
                    'Total project size ({file} bytes) exceeds limit: '
                    '{lim} bytes'.format(file=sz,
                                         lim=size_limit)
                )
        return True

    @observe('resources')
    def _fix_proj_res(self, change):
        before = change['old']
        after = change['new']
        if before in (None, Undefined):
            before = []
        if after in (None, Undefined):
            after = []
        for res in after:
            if res not in before and self not in res.project:
                res.project += [self]
        for res in before:
            if res not in after and self in res.project:
                res.project = [p for p in res.project
                               if p is not self]
        if len(set(after)) != len(after):
            post_post = []
            for r in after:
                if r not in post_post:
                    post_post += [r]
            self.resources = post_post

    def _upload_dirty(self, sync=False, verbose=True, tab_level=''):
        dirty = self._dirty
        if 'resources' in dirty:
            [r._upload(sync, verbose, tab_level) for r in self.resources]

    def _get_dirty_data(self, force=False, initialize=False):
        datadict = super(Project, self)._get_dirty_data(force)
        dirty = self._dirty_traits
        if 'public' in dirty or force:
            datadict['public'] = self.public
        if ('resources' in dirty or force) and not initialize:
            datadict['resourceUids'] = ','.join(
                (r._json['longUid'] for r in self.resources)
            )
        return datadict

    def _check_project_quota(self, verbose=True):
        if self.public:
            privacy = 'public'
        else:
            privacy = 'private'
        if verbose:
            print('Verifying your quota for ' + privacy + ' projects...')
        resp = Comms.get('api/check/quota?test=ProjectSteno3D')
        resp = resp['json'][privacy]
        if resp['quota'] == 'Unlimited':
            pass
        elif resp['count'] >= resp['quota']:
            raise Exception(QUOTA_REACHED.format(
                                priv=privacy,
                                num=resp['quota']
                            ))
        elif verbose and (resp['quota'] - resp['count'] - 1) < 4:
            print(QUOTA_IMPENDING.format(
                    remaining=resp['quota'] - resp['count'] - 1,
                    priv=privacy
                  ))
        if verbose and self.public:
            print('This PUBLIC project will be viewable by everyone.')

    @property
    def _url(self):
        if getattr(self, '_upload_data', None) is not None:
            return self._url_view_from_uid(self._upload_data['uid'])

    @property
    @needs_login
    def url(self):
        """steno3d.com url of project if uploaded"""
        if getattr(self, '_upload_data', None) is None:
            print('Project not uploaded: Please upload() '
                  'before accessing the URL.')
        return self._url

    @needs_login
    def plot(self):
        """Display the 3D representation of the content"""
        if getattr(self, '_upload_data', None) is None:
            print('Project not uploaded: Please upload() '
                  'before plotting.')
            return
        return plot(self._url)

    @classmethod
    def _build(cls, uid, copy=True, tab_level=''):
        print('Downloading project', end=': ')
        json = cls._json_from_uid(uid)
        title = '' if json['title'] is None else json['title']
        desc = '' if json['description'] is None else json['description']
        print(title)
        pub = False
        for a in json['access']:
            if a['user'] == 'Special:PUBLIC':
                pub = True
                break
        is_owner = Comms.user.username == json['owner']['uid']
        if copy is None:
            copy = not is_owner
        elif not copy and not is_owner:
            copy = True
        if copy:
            print('This is a copy of the {pub} project'.format(
                pub='PUBLIC' if pub else 'private'
            ))
        else:
            print('This is the original version of the {pub} project'.format(
                pub='PUBLIC' if pub else 'private'
            ))
            print('>> NOTE: Any changes you upload will overwrite the '
                  'project on steno3d.com')
            print('>> ', end='')
            if len(json['perspectiveUids']) > 0:
                print('and existing perspectives may be invalidated. ', end='')
            print('Please upload with caution.')

        proj = Project(
            public=pub,
            title=title,
            description=desc,
            resources=[]
        )
        for longuid in json['resourceUids']:
            res_string = longuid.split('Resource')[-1].split(':')[0]
            res_class = _REGISTRY[res_string]
            proj.resources += [res_class._build(
                src=longuid.split(':')[1],
                copy=copy,
                tab_level=tab_level + '    ',
                project=proj
            )]
        if not copy:
            proj._public_online = pub
            proj._upload_data = json
            proj._mark_clean()
        print('... Complete!')
        return proj

    @classmethod
    def from_omf(cls, omf_input):
        if isinstance(omf_input, six.string_types):
            from omf import OMFReader
            omf_input = OMFReader(omf_input)
        if not omf_input.__class__.__name__ == 'Project':
            raise ValueError('input must be omf file or Project')
        return cls._build_from_omf(omf_input)

    @classmethod
    def _build_from_omf(cls, omf_project):
        proj = Project(
            title=omf_project.name,
            description=omf_project.description,
            resources=[]
        )
        resource_map = {
            'PointSetElement': 'Point',
            'LineSetElement': 'Line',
            'SurfaceElement': 'Surface',
            'VolumeElement': 'Volume'
        }
        for elem in omf_project.elements:
            res_class = _REGISTRY[resource_map[elem.__class__.__name__]]
            proj.resources += [
                res_class._build_from_omf(elem, omf_project, proj)
            ]
        return proj


__all__ = ['Project']



"""teapot.py provides an example Steno3D project of a teapot"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import loads

from numpy import array

from .base import BaseExample, exampleproperty
from ..point import Mesh0D, Point
from ..project import Project
from ..surface import Mesh2D, Surface


class Teapot(BaseExample):
    """Teapot example

    This module provides example data for a teapot.
    """

    @exampleproperty
    def filenames(self):
        return ['teapot.json']

    @exampleproperty
    def _data(self):
        """teapot data read from json"""
        if getattr(self, '__data', None) is None:
            json_file = Teapot.fetch_data(filename='teapot.json',
                                          download_if_missing=False,
                                          verbose=False)
            with open(json_file, 'r') as f:
                self.__data = loads(f.read())
        return self.__data

    @exampleproperty
    def vertices(self):
        """n x 3 numpy array of teapot vertices"""
        return array(self._data['vertices'])

    @exampleproperty
    def triangles(self):
        """n x 3 numpy array of teapot triangle vertex indices"""
        return array(self._data['triangles'])

    @classmethod
    def get_project(self):
        """return teapot project"""
        proj = Project(
            title='Teapot',
            description='Project with surface and points at vertices'
        )
        Point(
            project=proj,
            mesh=Mesh0D(
                vertices=self.vertices
            ),
            title='Teapot Vertex Points'
        )
        Surface(
            project=proj,
            mesh=Mesh2D(
                vertices=self.vertices,
                triangles=self.triangles
            ),
            title='Teapot Surface'
        )
        return proj



#!/usr/bin/env python
'''
    steno3d: Client library for Steno3D & steno3d.com
'''

from distutils.core import setup
from setuptools import find_packages

CLASSIFIERS = [
'Development Status :: 4 - Beta',
'Programming Language :: Python',
'Topic :: Scientific/Engineering',
'Topic :: Scientific/Engineering :: Mathematics',
'Topic :: Scientific/Engineering :: Physics',
'Operating System :: Microsoft :: Windows',
'Operating System :: POSIX',
'Operating System :: Unix',
'Operating System :: MacOS',
'Natural Language :: English',
]

with open('README.rst') as f:
    LONG_DESCRIPTION = ''.join(f.readlines())

setup(
    name = 'steno3d',
    version = '0.2.14',
    packages = find_packages(),
    install_requires = ['future',
                        'numpy>=1.7',
                        'pypng',
                        'requests',
                        'six',
                        'traitlets>=4.1',
                       ],
    author = '3point Science',
    author_email = 'info@3ptscience.com',
    description = 'steno3d',
    long_description = LONG_DESCRIPTION,
    keywords = 'visualization',
    url = 'https://steno3d.com/',
    download_url = 'http://github.com/3ptscience/steno3dpy',
    classifiers=CLASSIFIERS,
    platforms = ['Windows', 'Linux', 'Solaris', 'Mac OS-X', 'Unix'],
    license='MIT License',
    use_2to3 = False,
)



"""airports.py provides an example Steno3D project of airports"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np

from .base import BaseExample, exampleproperty
from ..point import Mesh0D, Point
from ..project import Project

DEG2RAD = np.pi/180
FT2KM = 12*2.54/100/1000
RADIUS = 6371


class Airports(BaseExample):
    """Class containing components of airport project. Components can be
    viewed individually or copied into new resources or projects with
    get_resources() and get_project(), respectively.
    """

    @exampleproperty
    def filenames(self):
        """airport files"""
        return ['airports.dat', 'latitude.npy', 'longitude.npy',
                'altitude.npy', 'license.txt']

    @exampleproperty
    def datafile(self):
        """full path to airport data file"""
        return Airports.fetch_data(filename='airports.dat',
                                   download_if_missing=False,
                                   verbose=False)

    @exampleproperty
    def latitude(self):
        """Airport lat, degrees, from openflights.org"""
        return np.load(Airports.fetch_data(filename='latitude.npy',
                                           download_if_missing=False,
                                           verbose=False))

    @exampleproperty
    def longitude(self):
        """Airport lon, degrees, from openflights.org"""
        return np.load(Airports.fetch_data(filename='longitude.npy',
                                           download_if_missing=False,
                                           verbose=False))

    @exampleproperty
    def altitude(self):
        """Airport alt, km, from openflights.org"""
        return np.load(Airports.fetch_data(filename='altitude.npy',
                                           download_if_missing=False,
                                           verbose=False))

    @classmethod
    def get_project(self):
        """return airport points project"""
        proj = Project(
            title='Airport',
            description='Project with airport points'
        )
        Point(
            project=proj,
            mesh=Mesh0D(
                vertices=np.c_[self.geo_to_xyz(self.latitude,
                                               self.longitude,
                                               self.altitude)]
            ),
            title='Airport Points'
        )
        return proj

    @staticmethod
    def geo_to_xyz(lat, lon, alt):
        """function geo_to_xyz

        Inputs:
            lat: latitude, degrees
            lon: longitude, degrees
            alt: altitude, km

        Outputs:
            x, y, z: spatial coordiantes relative to the center of the earth

        Note:
            This function assumes a shpherical earth
        """

        lat *= DEG2RAD
        lon *= DEG2RAD
        x = (RADIUS + alt)*np.cos(lat)*np.cos(lon)
        y = (RADIUS + alt)*np.cos(lat)*np.sin(lon)
        z = (RADIUS + alt)*np.sin(lat)

        return x, y, z

    @staticmethod
    def read_airports_data(filename):
        """Extract latitude, longitude, and altitude from file"""
        lat = []  # Latitude
        lon = []  # Longitude
        alt = []  # Altitude
        with open(filename) as f:
            for line in f:
                data = line.rstrip().split(',')
                lat.append(float(data[6])*DEG2RAD)
                lon.append(float(data[7])*DEG2RAD)
                alt.append(float(data[8])*FT2KM)
        return np.array(lat), np.array(lon), np.array(alt)



"""options.py defines the base option classes for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
from traitlets import All, observe

from .traits import Bool, Color, Float, HasSteno3DTraits


class Options(HasSteno3DTraits):
    """Generic options for all steno3d resources"""

    @property
    def _json(self):
        """returns json representation of options"""
        opts_json = {}
        for key in self.trait_names():
            opts_json[key] = getattr(self, key)
        return dumps(opts_json)

    @observe(All)
    def _opt_defaulter(self, change):
        if change['new'] is None:
            owner = change['owner']
            name = change['name']
            setattr(owner, name, owner.traits()[name].default_value)


class ColorOptions(Options):
    """Options related to resource display color"""
    color = Color(
        help='Solid color',
        default_value='random',
        allow_none=True
    )
    opacity = Float(
        help='Opacity',
        default_value=1.,
        min=0.,
        max=1.,
        allow_none=True
    )


class MeshOptions(Options):
    """Options related to mesh display"""
    wireframe = Bool(
        help='Wireframe',
        default_value=False,
        allow_none=True
    )
